/*
**    TP ETI5 CPE Lyon
**    Copyright (C) 2010 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/



#include <glut_helper.hpp>

#include <GL/glew.h>
#include <GL/glut.h>
#include <string_helper.hpp>

#include <matrix3.hpp>


void setup_opengl(int argc,char *argv[],const cpe::navigator_tool& nav)
{
    glutInit (&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
    glutInitWindowPosition(0,0);
    glutInitWindowSize(nav.screen_size_x(),nav.screen_size_y());
    glutCreateWindow("OpenGL Window; CPE");

    glewInit();


    //OpenGL set
    glClearColor(1.0f,1.0f,1.0f,1.0f);
    glPointSize(5);
    glLineWidth(2);

    glShadeModel(GL_SMOOTH);
    glEnable(GL_COLOR_MATERIAL);
    //glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE );
    glEnable(GL_DEPTH_TEST);
    glDepthMask(GL_TRUE);


    glFrontFace(GL_CCW);//clockwise CW or CCW
    glDisable(GL_CULL_FACE);
    //glEnable(GL_CULL_FACE);

    glDisable(GL_BLEND);
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_FOG);

    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    glFrontFace( GL_CW ) ;
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);

    glEnable(GL_LINE_SMOOTH);
    glHint(GL_LINE_SMOOTH_HINT,GL_NICEST);







}



//keyboard
void keyboard_manager(unsigned char key,int x,int y,cpe::navigator_tool& nav)
{

  x=0;y=0;
  double dL=0.01*(1+10*abs(nav.dist_cam()));

  #define GLUT_SPACE 32


  switch(key)
  {
  case 'Q'://quit
      std::cout<<"\n[EXIT OK]\n\n"<<std::endl;
      exit(0);
      break;


  case 'z':
      nav.go_forward(dL);
      break;
  case 's':
      nav.go_forward(-dL);
      break;
  case 'q':
      nav.go_right(-dL);
      break;
  case 'd':
      nav.go_right(dL);
      break;
  case 'r':
      nav.go_up(dL);
      break;
  case 'f':
      nav.go_up(-dL);
      break;





  case 'W'://wireframe
      glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
      break;
  case 'F'://Filled
      glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
      break;

  }

  glutPostRedisplay ();
}


void motion_manager(int x, int y,cpe::navigator_tool& nav)
{

    if ( (nav.left_button()) )
        nav.trackball_rotate(x,y);
    else if ( (nav.right_button()))
        nav.trackball_zoom(y);

    glutPostRedisplay ();
}

void mouse_manager(int button,int action,int x,int y,cpe::navigator_tool& nav)
{
    nav.x_previous()=x;
    nav.y_previous()=y;

    if (button == GLUT_LEFT_BUTTON && action == GLUT_DOWN)
        nav.left_button()=true;
    if (button == GLUT_LEFT_BUTTON && action == GLUT_UP)
        nav.left_button()=false;
    if (button == GLUT_RIGHT_BUTTON && action == GLUT_DOWN)
        nav.right_button()=true;
    if (button == GLUT_RIGHT_BUTTON && action == GLUT_UP)
        nav.right_button()=false;

    glutPostRedisplay ();
}

void reshape_manager(int width,int height,cpe::navigator_tool& nav)
{
    nav.screen_size_x()=width;nav.screen_size_y()=height;glViewport(0,0, width, height);
}

void print_help_start()
{
    std::cout<<"**RUN**\n"<<std::endl;
    std::cout<<"Navigation with keys:\n\n"<<std::endl;
    std::cout<<"==========================\n"<<std::endl;
    std::cout<<"Space Displacement"<<std::endl;
    std::cout<<"  z    r"<<std::endl;
    std::cout<<"q s d  f\n"<<std::endl;
    std::cout<<"and mouse with drag & drop\n"<<std::endl;
    std::cout<<"W: wireframe mode"<<std::endl;
    std::cout<<"F: filled mode"<<std::endl;
    std::cout<<"Q: exit"<<std::endl;
    std::cout<<"==========================\n"<<std::endl;
}






void display_manager_begin(cpe::navigator_tool& nav)
{

    //vieport (all screen)
    glViewport (0, 0, nav.screen_size_x(),nav.screen_size_y());
    glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    //Projection
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity ();
    std::vector<double> proj_param=nav.projection_param();
    glFrustum(proj_param[0], -proj_param[0], proj_param[1], -proj_param[1], proj_param[2], proj_param[3]);


    //Modelview
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glTranslated(0,0,static_cast<GLdouble>(nav.dist_cam())); //zoom
    std::vector<double> m=nav.camera_orientation().to_4d();
    glMultMatrixd(&m[0]); //rotation

    //translation
    cpe::v3 tr=nav.camera_translation();
    glTranslated(static_cast<GLdouble>(tr[0]),static_cast<GLdouble>(tr[1]),static_cast<GLdouble>(tr[2]));

}


void display_manager_end(cpe::navigator_tool& nav,unsigned int& counter)
{

    //draw_pointer
    draw_pointer(nav);

    //draw orientation (clear DEPTH-BUFFER)
    draw_orientation(nav);

    //update fps
    nav.increase_frame_number();
    nav.time()=glutGet(GLUT_ELAPSED_TIME);
    if(nav.delta_time()>1000)
    {nav.update_fps();}

    //draw fps
    draw_fps(nav);

    //error OpenGL_drawer_2
    GLenum errCode;
    const GLubyte *errString;
    if((errCode=glGetError()) != GL_NO_ERROR)
    {
        errString = gluErrorString(errCode);
        std::cout<<"OPENGL ERROR : "<<errString<<std::endl;
        throw std::string("OpenGL Error");
    }

    counter++;


    //redraw
    glFlush ();
    glutSwapBuffers ();

}


void draw_pointer(cpe::navigator_tool& nav)
{

    glPushMatrix();
    glLoadIdentity();

    glTranslatef(0,0,GLfloat(nav.dist_cam())); //zoom
    std::vector<double> m=nav.camera_orientation().to_4d();
    glMultMatrixd(&m[0]); //rotation

    //local basis
    double L=0.01;

    glDisable(GL_LIGHTING);
    glBegin(GL_LINES);
    glColor3d(1,0,0);
    glVertex3d(-L,0,0);
    glVertex3d(L,0,0);
    glColor3d(0,1,0);
    glVertex3d(0,-L,0);
    glVertex3d(0,L,0);
    glColor3d(0,0,1);
    glVertex3d(0,0,-L);
    glVertex3d(0,0,L);
    glEnd();

    glPopMatrix();

}

void draw_orientation(cpe::navigator_tool& nav)
{
    glDisable(GL_LIGHTING);
    //rotation
    glClear(GL_DEPTH_BUFFER_BIT);
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity ();
    glOrtho(-1,1,1,-1,0.1,10.0);
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    glTranslated(0.78,0.78,-1);
    std::vector<double> m=nav.camera_orientation().to_4d();
    // rotation
    glMultMatrixd(&m[0]);

    double L=0.2;
    glBegin(GL_LINES);
    glColor3d(1,0,0);
    glVertex3d(0,0,0);
    glVertex3d(L,0,0);
    glColor3d(0,1,0);
    glVertex3d(0,0,0);
    glVertex3d(0,L,0);
    glColor3d(0,0,1);
    glVertex3d(0,0,0);
    glVertex3d(0,0,L);
    glEnd();
    //pop matrices
    glPopMatrix();
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);

}


void draw_fps(cpe::navigator_tool& nav)
{
    glDisable(GL_DEPTH_TEST);

    int int_fps=int(nav.fps());
    int deci_fps=int(10*(nav.fps()-int_fps));
    double truncated_fps=int_fps+double(deci_fps)/10.0;

    std::string fps=(cpe::string_converter::to_string(truncated_fps)+" fps");

    glPushMatrix();
    glLoadIdentity();
    glRasterPos3f(-0.5f,-0.5f,-1.4f);
    for (unsigned int k=0,N=fps.length();k<N;++k)
        glutBitmapCharacter(GLUT_BITMAP_TIMES_ROMAN_24,fps[k]);
    glPopMatrix();

    glEnable(GL_DEPTH_TEST);
}

