// animated_mesh.cc

#include <iostream>
#include <fstream>
#include <sstream>

#define GLEW_STATIC 1
#include <GL/glew.h>
#include <GL/glut.h>

#include "gl-utils.hpp"

#include "animated_mesh.hpp"
#include <mesh.hpp>
#include <opengl_drawer.hpp>

void
AnimatedMesh::load (const std::string &filename)
{

  unsigned      num_vertex, num_triangles;
  std::string   line;
  std::ifstream file (filename.c_str());

  if(!file.good())
  {throw std::string("cannot open file in AnimatedMesh::load("+filename+")");}

  vertices.clear ();
  indices.clear ();
  std::getline (file, line); // FIXME, should compare with "AOFF"
  std::getline (file, line);
  std::stringstream tokens (line);
  tokens >> num_vertex >> num_triangles;

  vertices.resize (num_vertex);
  tex_coords.resize (num_vertex * 2);
  for (unsigned i = 0; i < num_vertex; i++)
    {
      std::getline (file, line);
      if (line[0] == '#') continue;
      std::stringstream tokens2 (line);
      for (int j = 0; j < 4; j++)
        {
          tokens2 >> vertices[i].parts[j].bone;
          tokens2 >> vertices[i].parts[j].weight;
          tokens2 >> vertices[i].parts[j].position.x();
          tokens2 >> vertices[i].parts[j].position.y();
          tokens2 >> vertices[i].parts[j].position.z();
        }
      tokens2 >> tex_coords[i * 2 + 0];
      tokens2 >> tex_coords[i * 2 + 1];
    }
  for (unsigned i = 0; i < num_triangles; i++)
    {
      GLuint dummy, index;
      std::getline (file, line);
      if (line[0] == '#') continue;
      std::stringstream tokens2 (line);
      tokens2 >> dummy;
      tokens2 >> index; indices.push_back (index);
      tokens2 >> index; indices.push_back (index);
      tokens2 >> index; indices.push_back (index);
    }



  // Fill the attribs
  for (unsigned i = 0; i < vertices.size (); i++)
    {
      for (unsigned j = 0; j < 4; j++)
        {
          vertex_attrib0123.push_back (vertices[i].parts[j].position.x());
          vertex_attrib0123.push_back (vertices[i].parts[j].position.y());
          vertex_attrib0123.push_back (vertices[i].parts[j].position.z());
          vertex_attrib0123.push_back (vertices[i].parts[j].weight);
          vertex_attrib4.push_back (vertices[i].parts[j].bone);
        }
    }


}

void AnimatedMesh::setup_shader(const std::string& vertex_shader,const std::string& fragment_shader)
{

  program = read_shader (vertex_shader, fragment_shader);
  glUniform1i (get_uni_loc (program, "diffuse_map"), 0);


  weight0_index = glGetAttribLocation (program, "weight0");
  weight1_index = glGetAttribLocation (program, "weight1");
  weight2_index = glGetAttribLocation (program, "weight2");
  weight3_index = glGetAttribLocation (program, "weight3");
  bones_index = glGetAttribLocation (program, "bones");



  glGenBuffers (1, &vbo_weights);
  glBindBuffer (GL_ARRAY_BUFFER, vbo_weights);
  glBufferData (GL_ARRAY_BUFFER, 16 * vertices.size() * sizeof (GLfloat), &vertex_attrib0123[0], GL_STATIC_DRAW);
  glVertexAttribPointer (weight0_index, 4, GL_FLOAT, GL_FALSE, 16 * sizeof (GLfloat), BUFFER_OFFSET (0));
  glVertexAttribPointer (weight1_index, 4, GL_FLOAT, GL_FALSE, 16 * sizeof (GLfloat), BUFFER_OFFSET (4 * sizeof (GLfloat)));
  glVertexAttribPointer (weight2_index, 4, GL_FLOAT, GL_FALSE, 16 * sizeof (GLfloat), BUFFER_OFFSET (8 * sizeof (GLfloat)));
  glVertexAttribPointer (weight3_index, 4, GL_FLOAT, GL_FALSE, 16 * sizeof (GLfloat), BUFFER_OFFSET (12 * sizeof (GLfloat)));

  glGenBuffers (1, &vbo_bones);
  glBindBuffer (GL_ARRAY_BUFFER, vbo_bones);
  glBufferData (GL_ARRAY_BUFFER, 4 * vertices.size() * sizeof (GLfloat), &vertex_attrib4[0], GL_STATIC_DRAW);
  glVertexAttribPointer (bones_index, 4, GL_FLOAT, GL_FALSE, 0, BUFFER_OFFSET (0));

  glGenBuffers (1, &vbo_textures);
  glBindBuffer (GL_ARRAY_BUFFER, vbo_textures);
  glBufferData (GL_ARRAY_BUFFER, 2 * vertices.size() * sizeof (GLfloat), &tex_coords[0], GL_STATIC_DRAW);
  glTexCoordPointer (2, GL_FLOAT, 0, BUFFER_OFFSET (0));
  glBindBuffer (GL_ARRAY_BUFFER, 0);

  glGenBuffers (1, &vbo_indices);
  glBindBuffer (GL_ELEMENT_ARRAY_BUFFER, vbo_indices);
  glBufferData (GL_ELEMENT_ARRAY_BUFFER, indices.size () * sizeof (GLuint), &indices[0], GL_STATIC_DRAW);

}

//skinning sur CPU
void AnimatedMesh::draw_cpu(const std::vector<cpe::v3> &skel_pos,
                            const std::vector<cpe::quaternion> &skel_orient,
                     bool update)
{

    if(!update)
        return;

    //on créé une nouvelle classe de maillage classique (pas optimal en temps)
    cpe::mesh mesh;
    //on copie la connectivité qui ne varie pas
    mesh.get_connectivity()=indices;
    //on copie les coordonnées de textures
    std::vector<double> tex_tmp;
    for(unsigned int k=0,N=tex_coords.size();k<N;++k)
        tex_tmp.push_back(tex_coords[k]);
    mesh.get_texture()=tex_tmp;

    //*********************************************
    //BOUCLE DE DEFORMATION PAR SKINNING
    //*********************************************
    for(unsigned int k_vertex=0,N_vertex=vertices.size();k_vertex<N_vertex;++k_vertex)
    {
        //positions concatenees
        cpe::v3 p0 (vertex_attrib0123[4*(4*k_vertex+0)+0],vertex_attrib0123[4*(4*k_vertex+0)+1],vertex_attrib0123[4*(4*k_vertex+0)+2]);
        //poids the skinning
        double w0 = vertex_attrib0123[4*(4*k_vertex+0)+3];
        //indice de l'os courant
        int b0=vertex_attrib4[4*k_vertex+0];
        //quaternion courant
        cpe::quaternion q0=skel_orient[b0];
        //position global du joint
        cpe::v3 tr0=skel_pos[b0];
        //position du sommet centre sur le joint
        cpe::v3 relative0=q0*p0;

        //calcul de la nouvelle position du sommet (à modifier!)
        cpe::v3 y = tr0+relative0;

        //ajout du nouveau sommet dans la classe de maillage
        mesh.add_vertex(y);
    }

    //affichage standard d'un maillage
    cpe::opengl_drawer::draw(mesh,mesh.normal_polygon());
}


//skinning sur GPU
void
        AnimatedMesh::draw (const std::vector<cpe::v3> &skel_pos,
                            const std::vector<cpe::quaternion> &skel_orient,
                     bool update)
{
  glUseProgram (program);


  if (update)
  {
      //convert double to GLFloat
      std::vector <GLfloat> pos;
      std::vector <GLfloat> ori;
      for(unsigned int k=0;k<skel_pos.size();++k)
          for(unsigned int k_dim=0;k_dim<3;++k_dim)
              pos.push_back(skel_pos[k][k_dim]);
      for(unsigned int k=0;k<skel_orient.size();++k)
          for(unsigned int k_dim=0;k_dim<4;++k_dim)
              ori.push_back(skel_orient[k][k_dim]);


      glUniform3fv (get_uni_loc (program, "position[0]"), skel_pos.size (), &pos[0]);
      glUniform4fv (get_uni_loc (program, "orientation[0]"), skel_orient.size (), &ori[0]);
  }

  glEnableVertexAttribArray (weight0_index);
  glEnableVertexAttribArray (weight1_index);
  glEnableVertexAttribArray (weight2_index);
  glEnableVertexAttribArray (weight3_index);
  glEnableVertexAttribArray (bones_index);
  glEnableClientState (GL_TEXTURE_COORD_ARRAY);

  glDrawElements (GL_TRIANGLES, indices.size (), GL_UNSIGNED_INT, 0);

  glDisableVertexAttribArray (weight0_index);
  glDisableVertexAttribArray (weight1_index);
  glDisableVertexAttribArray (weight2_index);
  glDisableVertexAttribArray (weight3_index);
  glDisableVertexAttribArray (bones_index);
  glDisableClientState (GL_TEXTURE_COORD_ARRAY);

  glUseProgram (0);
}

