
/** TP ETI5 - CPE Lyon - 2010-2011 */



#include <v3.hpp>
#include <mesh.hpp>
#include <mesh_io_off.hpp>
#include <opengl_drawer.hpp>
#include <matrix3.hpp>
#include <navigator_tool.hpp>
#include <string_helper.hpp>
#include <glut_helper.hpp>
#include <texture.hpp>

#include <GL/glew.h>
#include <GL/glut.h>

#include <stdlib.h>
#include <iostream>
#include <cmath>
#include <string>
#include <errno.h>
#include <vector>

#include <animation.hpp>
#include <animated_mesh.hpp>
#include <gl-utils.hpp>


//draw the model
int draw_scene();
//load the model
int load_model();
//compute a random number
double get_rand();



//glut functions
int setup_callback();
static void display_callback();
static void reshape_callback (int width, int height);
static void keyboard_callback(unsigned char key,int x,int y);
static void idle_callback();
static void motion_callback (int x, int y);
static void mouse_callback (int button, int action, int x, int y);

static void
interpolate_skeletons (const std::vector<Joint> &in1,
                       const std::vector<Joint> &in2,
                       float                     t);

using namespace cpe;




//GLOBAL VARIABLES
//************************************//
unsigned int counter=0;
navigator_tool nav;



//skinning
#define FRAMERATE 24.0
Animation               animation;
AnimatedMesh            mesh_skinning;

std::vector<cpe::v3>       skel_position;
std::vector<cpe::quaternion> skel_orientation;
bool is_GPU=false;




//load the model
int load_model()
{

    try
    {
        opengl_drawer::generate_texture(1,"data_skinning/texture/texture.png");
        animation.load ("data_skinning/anim/bind_pose.animation");
        mesh_skinning.load ("data_skinning/mesh/body.basemesh");
        if(is_GPU)
            mesh_skinning.setup_shader("shaders/skinner.vert", "shaders/skinner.frag");

        skel_position.resize (animation.bones[0].size ());
        skel_orientation.resize (animation.bones[0].size ());

    }
    catch(std::string e)
    {
        std::cout<<"Exception: "<<e<<std::endl;
    }

    return 0;
}


//Draw the scene
int draw_scene()
{
    glEnable(GL_POLYGON_OFFSET_FILL);
    glPolygonOffset(5,5);


    static double current_time = 0;
    static double last_time = 0;
    static double info_last_time = 0;

    current_time = glutGet (GLUT_ELAPSED_TIME) / 1000.0;
    info_last_time += current_time - last_time;
    last_time = current_time;
    if (info_last_time >= 1.0 / FRAMERATE)
    {
        info_last_time -= 1.0 / FRAMERATE;
    }

    /////////////////////////////////////////
    /////////////////////////////////////////
    // TO DO: interpoler les poses
    /////////////////////////////////////////
    /////////////////////////////////////////
    interpolate_skeletons (animation.bones[0],animation.bones[1],0);

    glEnable(GL_LIGHTING);
    glColor3d(0.7,0.7,0.7);
    //glEnable(GL_TEXTURE_2D);

    if(is_GPU)
    {
        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
        mesh_skinning.draw (skel_position, skel_orientation, true);
    }
    else
        mesh_skinning.draw_cpu (skel_position, skel_orientation, true);

    glDisable(GL_TEXTURE_2D);
    glDisable(GL_LIGHTING);


    return 0;
}




static void
interpolate_skeletons (const std::vector<Joint> &joint_vector_time1,
                       const std::vector<Joint> &joint_vector_time2,
                       float                     t)
{
    /////////////////////////////////////////
    /////////////////////////////////////////
    // TO DO: interpoler lineairement les positions, et avec SLERP pour les orientations
    /////////////////////////////////////////
    /////////////////////////////////////////
    for(unsigned int k=0;k<skel_position.size();++k)
    {
        skel_position[k]=joint_vector_time1[k].position;
        skel_orientation[k]=joint_vector_time1[k].orientation;
    }

}











int main(int argc, char *argv[])
{
    std::cout<<"**hello**\n"<<std::endl;
    std::cout<<"run "<<argv[0]<<" with "<<argc-1<<" parameters ... \n"<<std::endl;



  // ************* //
  // start OpenGL
  // ************* //

  //window size
  nav.screen_size_x()=800;
  nav.screen_size_y()=800;



  //init opengl & glut
  setup_opengl(argc,argv,nav);
  //setup callback functions
  setup_callback();





  // ************* //
  // RUN
  // ************* //
  print_help_start();

  //do start once
  load_model();

  //glut loop
  glutMainLoop ();


  return 0;
}



// *******************//
// Helper
// *******************//

//just a rand number \in [0,1]
double get_rand()
{
  int precision=50000;
  return double(rand()%precision)/double(precision);
}






//*******************//
//GLUT STUFF
//*******************//


//glut functions
int setup_callback()
{

    //callback
    glutDisplayFunc (display_callback);
    glutReshapeFunc (reshape_callback);
    glutKeyboardFunc(keyboard_callback);
    glutIdleFunc(idle_callback);
    glutMotionFunc (motion_callback);
    glutMouseFunc (mouse_callback);

    return 0;
}



//display
static void display_callback()
{
    display_manager_begin(nav);
    draw_scene();
    display_manager_end(nav,counter);
}



static void reshape_callback (int width, int height) {reshape_manager(width,height,nav);}
static void idle_callback() {glutPostRedisplay ();}
static void keyboard_callback(unsigned char key,int x,int y) {keyboard_manager(key,x,y,nav);}
static void motion_callback (int x, int y) {motion_manager(x,y,nav);}
static void mouse_callback (int button, int action, int x, int y) {mouse_manager(button,action,x,y,nav);}












